"""
Script containing various additional helper functions
"""
import pkgutil
import os
import sys
from bs4 import BeautifulSoup
import requests
import requirements
import re
import subprocess
import yaml
import json

def extract_keys(nested_dict):
	"""
	This function is used in order to parse the.json output generated by executing conda search <package-name>
	for potential install candidates.

	Parameters
	----------
	nested_dict : dictionary
		.json file

	Returns
	-------
	key_lst: list
		list containing only the keys (the potential installation candidates)
	"""
	key_lst = []
	for key, value in nested_dict.items():
		key_lst.append(key)
		if isinstance(value, dict):
			extract_keys(value)
	return key_lst

def dir_path(str):
	"""
	Check if path leads to a valid directory

	Parameters
	----------
	str : str
		path to directory

	Returns
	-------
	str
		input if directory exists
	"""
	if os.path.isdir(str):
		return str
	else:
		raise NotADirectoryError(str)

def replace_names(module_lst):
	"""
	Load name-assignment.txt and compare against extracted import modules
	If a match is found, replace with value. (which is the according package name for conda/pip)
	As for now this is the way to approach the discrepancy between
	module names as and package names as described in:
	https://stackoverflow.com/questions/38733220/difference-between-scikit-learn-and-sklearn

	Parameters
	----------
	module_lst : list
		Contains module names extracted from import statements

	Returns
	-------
	module_lst
	"""

	name_assign_dict = {}
	module_names_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'module-names.txt')
	with open(module_names_path, 'r') as f:
		for line in f:
			(key, val) = line.split()
			name_assign_dict[key] = val

	for i, elem in enumerate(module_lst):
		if elem in name_assign_dict:
			module_lst[i] = name_assign_dict[elem]

	return module_lst

def display_options(numbered_lst, elem):
	"""
	Displays command options and selection choices.
	

	Parameters
	----------
	numbered_lst : list
		enumerated installation candidates for elem
	elem : str
		package name to be installed
	"""
	print(f"I found several options for {elem}")
	print(f"Please select a choice: ") 
	print("'s' for skip channel")
	print("'c' for cancel search for this package")
	print("'n' for adding a custom package name instead")
	print("'q' for quit")

	for option in numbered_lst:
		print(option)

def enumerate_installation_candidates(installation_candidates):
	"""
	Parameters
	----------
	installation_candidates : list

	Returns
	-------
	install_candidates_numbered: list
	"""
	install_candidates_numbered = [f"[{i}] {elem}" for i, elem in enumerate(installation_candidates)]
	return install_candidates_numbered

def start_menu():
	"""
	Displays menu at execution start.

	Returns
	-------
	name: str
		if provided custom environment name, else "test-env"
	channel_lst: lst
		if provided, contains additional conda-channels to install from
	"""
	print("Hello :)")
	#What kind of checks to add here?
	name = input("Please provide a name for the new environment: [e.g testenv, test-env, TestEnv] ")
	if not name:
		name = "test-env"
	choice = input("Do you want to specify any additional conda-channels to use? [y/n]")
	if choice.casefold() == "y".casefold():
		channel_lst = input("Please enter the conda-channels you want to use, separated by ',' ")
		channel_lst = channel_lst.split(',')
	else:
		channel_lst = None

	return name, channel_lst

def menu(num_installation_candidates, elem):
	"""
	Main menu displayed at every action taken.	

	Parameters
	----------
	num_installation_candidates : list
		enumerated installation candidates
	elem : str
		package to be installed

	Note: I think the current return can be done better.
	(Something for more extra time) 

	Returns
	-------
	choice: string | int
		choice depending
	"""
	display_options(num_installation_candidates, elem)
	choice = input("Choice :")
	if choice.casefold() == "q".casefold():
		sys.exit("Bye :)")

	# In case you choose a command..
	if choice.isalpha():
		return choice.lower()

	elif not re.match("\d{1,3}", choice):
			print("Invalid input")
			# Is this ok or dangerous?
			display_options(num_installation_candidates,elem)
	else:
		if int(choice) > len(num_installation_candidates):
			print("Invalid number")
			display_options(num_installation_candidates,elem)

	#..else your choice was an installation candidate..
	return int(choice)

def clean_list(enumerated_lst):
	"""
	Parameters
	----------
	enumerated_lst: list
		enumerated installation candidates

	Returns
	-------
	cleaned_lst: list
		cleans up weird naming within installation candidates
	"""
	cleaned_lst = [re.sub(r'(\[\d+\]\s)', '', elem) for elem in enumerated_lst]
	return cleaned_lst

def choose_custom_name(elem):
	"""
	Insert custom package name.
	Example: cv2 is extracted from the imports and the tool has no chance to know 
	that the package name is 'opencv'. If you know this right away, you can add this here.

	Note: Part of this is mitigated by using replace_names(). If you know more assignments,
	feel free to add them to module-names.txt
	
	Parameters
	----------
	elem : str
		chosen installation candidate

	Returns
	-------
	name: str
		custom name for package
	"""
	name = input(f"Please provide an alternative name for {elem} :")
	#TODO: checks and tests
	return name

def parse_requirements(src_rep):
	"""
	If the repository contains a requirements.txt we happily parse it

	Parameters
	----------
	src_rep : str | list
		repository to be parsed by findimports
		
	Returns
	-------
	pip_packages: list
		extracted package names
	"""
	pip_packages = []
	try:
		with open(os.path.join(src_rep,'requirements.txt'), 'r') as f:
			for req in requirements.parse(f):
				pip_packages.append(req.name)
			
			print("finished parsing requirements.txt")

		return pip_packages

	except FileNotFoundError:
		print("Could not find requirements.txt within specified src directory")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import subprocess
import tqdm
import argparse
import json
import re
import yaml
from .core_functions import determine_disjunct_module_names, search_conda_channel, create_env, search_pip
from .helpers import clean_list, dir_path, start_menu, parse_requirements,replace_names


def main():
	parser = argparse.ArgumentParser(description='''
													Extracts import statements from a python repository and returns a list of additionally added modules 
													(Future: creates a valid environment.yml file that can later be installed via conda and pip
												''')

	parser.add_argument('-s', '--src-repository', required=True, type=dir_path,
		help='''
				Provide a valid path to the source-code you want to parse e.g /path/to/my/src
			''')

	parser.add_argument('-l', '--leave-me-alone', action='store_true',default=False, 
		help='''
				Hides the decision query for each package/channel, sparing you a bit of interaction. 
				Good for lazy mode or for irritable moods.
				Be aware that it takes definitely longer tho.
			''')
	
	args = parser.parse_args()
	src_rep = args.src_repository
	leave_me_alone = args.leave_me_alone

	#check if the repository contains a requirements.txt
	pip_packages = parse_requirements(src_rep)

	#we might not need this due to default argument
	#leave_me_alone = True if leave_me_alone else False

	channel_lst = []
	name, additional_channel_lst = start_menu()
	default_channel_lst = ["conda-forge", "anaconda"]

	if additional_channel_lst:
		channel_lst = additional_channel_lst + default_channel_lst
	else:
		channel_lst = default_channel_lst

	print(f"channel_lst : {channel_lst}")


	# disjunct_modules, same_modules = determine_added_modules(src_rep,python_version)
	disjunct_modules = determine_disjunct_module_names(src_rep)

	disjunct_modules = replace_names(disjunct_modules)

	if pip_packages:
		disjunct_modules = list(set(disjunct_modules).union(pip_packages))

	print(f"Disjunct modules: {disjunct_modules}")


	# For each module-name, determine installation candidates (if available)
	# for now only one consider splitting for conda-forge/anaconda.

	conda_lst = []
	pip_lst = []
	module_not_found_lst = []

	for i,elem in enumerate(disjunct_modules):
		count = len(channel_lst)

		for channel in channel_lst:
			pname, skip = search_conda_channel(channel, elem, leave_me_alone)
			if pname is not None:
				conda_lst.append(pname)
				break
			else:
				if count > 0 and skip is True:
					print("Cancel search. Skip to next channel...")
					count -= count
					continue #break
				if count <= 0 or skip is False:
					module_not_found_lst.append(elem)
	
	module_not_found_lst = set(module_not_found_lst)

	
	for i, elem in enumerate(module_not_found_lst):
		pname, skip= search_pip(elem, leave_me_alone)
		if pname is not None:
			pip_lst.append(pname)
			continue
		else:
			if skip is True:
				print("Cancel search. Skip to next package...")
				continue
			else:
				print(f"{elem} not found within pip.")

	conda_lst = clean_list(conda_lst)
	if pip_lst:
		pip_lst = clean_list(pip_lst)

	print(f"conda candidates : {conda_lst}")
	print(f"pip candidates : {pip_lst}")
	print(f"the rest : {module_not_found_lst}")

	create_env(conda_lst,pip_lst,name,channel_lst)

if __name__ == "__main__":
	main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
This script needs to be executed only when a new python version has appeared.
chmod +x has been set

Execution:
$ ./generate_python_module_list.py
"""

from sys import version
from bs4 import BeautifulSoup
import requests
import re
import pickle
import argparse

def write_modules_list(python_version=3.10):
	"""
	This script parses https://docs.python.org/3.10/py-modindex.html# and creates a static list containing all the module names. 
	This list is used by determine_disjunct_module_names(src_rep) to compare against external modules within a given repository.

	Parameters
	----------
	python_version : str
		major+minor version e.g 3.7, 1.9, 3.10
	"""	
	try:
		page = requests.get(f'https://docs.python.org/{python_version}/py-modindex.html#')
		page.raise_for_status()

	except requests.exceptions.HTTPError as err:
		raise SystemExit(err)
	except requests.exceptions.RequestException as err:
		#catastrophic error. bail.
		raise SystemExit(err)

	content = page.text

	soup = BeautifulSoup(content, features="html.parser")
	#print(soup.prettify())

	#Fancy regex that emerged _before reading the docs of bs4 ^^
	#<code class="xref">\s*([\d\w_]+)\s*<\/code>
	pattern = soup.find_all("code", "xref")

	modules_lst = []
	for i, elem in enumerate(pattern):
		#print(f"element :{elem}, i : {i}")
		#This regex filters the module name out of the embeddings
		match = re.search(r'(?P<code0><code class="xref">)(?P<modulename>\s*([\d\w_.]+)\s*)(?P<code1><\/code>)', str(elem))
		modulename = match.group('modulename')
		modules_lst.append(modulename)

	#with open(f'python{python_version}-modules.txt', 'w') as f:
	#	modules_lst = map(lambda x:x+'\n', modules_lst)
	#	f.writelines(modules_lst)

	# append package name
	# only include the module name, e.g only skimage instead of skimage.filters
	modules_lst = list(set([ x.split('.')[0] for x in modules_lst]))
	pickle.dump(modules_lst, open('python_module_index.pickle', 'wb'))

if __name__ == "__main__":
	
	parser = argparse.ArgumentParser(description='''
								This script parses https://docs.python.org/3.10/py-modindex.html# 
								and creates a static list containing all the module names. 
												''')

	parser.add_argument('-v', '--version-number', required=True, type=str,
		help='''
				Provide a valid python version number. 
					Format: major+minor, e.g 3.7, 3.8, 3.10
			''')
	
	args = parser.parse_args()
	v_num = args.version_number

	write_modules_list(v_num)



"""
Contains code no longer used but kept for review/further reuse
--- imports need to be re-added ---
"""

def determine_disjuct_modules_alternative(src_rep):
	"""
		Potentially get rid of determine_added_modules and get_modules_lst()
	"""
	findimports_output = subprocess.check_output(['findimports', src_rep])
	findimports_output = findimports_output.decode('utf-8').splitlines()

	custom_modules_lst = []
	for i, elem in enumerate(findimports_output):
		if ':' in elem:
			continue
		elem = elem.rstrip('\n').split('.',1)[0].strip()
		#print(f" element : {elem}")
		custom_modules_lst.append(elem)

	custom_modules_lst = set(custom_modules_lst)

	#beautify this 
	disjunct_modules = []
	for i, elem in enumerate(custom_modules_lst):
		if elem in sys.modules:
			continue
		else:
			disjunct_modules.append(elem)
	
	return disjunct_modules

def determine_added_modules(src_rep, python_version):
	"""
	Determine overlapping and disjunct modules between the ones shipped by the specified python version and the ones found in the source repository specified.
	For now we rely on findimports providing the needed dependency tree within the source repository.
  	Links to tool : https://github.com/mgedmin/findimports, https://pypi.org/project/findimports/
	"""
	
	python_modules_lst = get_modules_list(python_version)

	findimports_output = subprocess.check_output(['findimports', src_rep])
	findimports_output = findimports_output.decode('utf-8').splitlines()

	custom_modules_lst = []
	for i, elem in enumerate(findimports_output):
		if ':' in elem:
			continue
		elem = elem.rstrip('\n').split('.',1)[0].strip()
		#print(f" element : {elem}")
		custom_modules_lst.append(elem)

	custom_modules_lst = set(custom_modules_lst)

	not_common = [val for val in custom_modules_lst if val not in python_modules_lst]
	common = [val for val in custom_modules_lst if val in python_modules_lst]

	return not_common, common


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gitcrawl",
    version="0.1.2",
    author="Neda Sultova",
    author_email="n.sultova@hzdr.de",
    description="A small package for creating os-agnostic conda environment.yml files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nsultova/gitcrawl",
    packages=setuptools.find_packages(),
    package_data={"gitcrawl": ["python_module_index.pickle", "module-names.txt"]},
    install_requires=[
		"beautifulsoup4",
		"tqdm",
		"requests",
		"PyYAML",
		"requirements-parser",
		"pip-search",
		"findimports"
    ],
    entry_points={
        'console_scripts':[
            'gitcrawl=gitcrawl:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
This script contains functions needed for the core functionality of gitcrawl
"""
import os
import sys
import re
import subprocess
import yaml
import json
from tqdm import tqdm
import time
import pickle
from .helpers import extract_keys, enumerate_installation_candidates, menu, choose_custom_name

def findimports_to_dict(src_rep):
	"""
	This function executes 

	Parameters
	----------
	src_rep: str | list
		repository to be parsed by findimports

	Returns
	-------
	dict: 
		module-names found 
		
	"""
	findimports = subprocess.run(
		['findimports'] + src_rep, capture_output=True)

	output = {}
	current_line = ''
	for line in findimports.stdout.decode('utf-8').splitlines():
		stripped = line.strip()
		# omit empty lines
		if len(stripped) == 0:
			continue
		# line has no added spaces in front. Thus it must be a directory.
		if len(stripped) == len(line):
			# remove the double colon at the end
			current_line = stripped[:-1]
			output[current_line] = []
			continue
		# append package name
		# only include the module name, e.g only skimage instead of skimage.filters
		output[current_line].append(stripped.split('.')[0])

	output = {k: set(output[k]) for k in output}

	return output

def determine_disjunct_module_names(src_rep):
	"""
	Filter the modules found by findimports in such a way that returned modules meet following criteria:
	* the module should not be part of the python installation running within this environment
	* the module should be external (exclude self-written modules)
	
	Notes:
		arising problem: E.g pickle which ships with each python installation but is nowhere to be found in sys.modules if not explicitly loaded
		hotfix: use python_module_index which is generated by generate_python_module_list.py 

	Parameters
	----------
	src_rep : str or list
		repository to be parsed by findimports
		
	Returns
	-------
	list:
		list containing unique module-names meeting above criteria
	"""
	# findimports only allows us to find either all imports or only the internally used ones.
	# but we need only the external ones
	all_modules = findimports_to_dict([src_rep])
	internal_modules = findimports_to_dict(['-N', src_rep])
	external_modules = {k: [ module for module in all_modules[k] if module not in internal_modules[k] ] for k in all_modules}

	external_modules_list = []
	for values in external_modules.values():
		for v in values:
			external_modules_list.append(v)
	custom_modules_lst = set(external_modules_list)

	# Filter out python internal modules (hotfix)
	python_module_index_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'python_module_index.pickle')
	python_modules = pickle.load(open(python_module_index_path, 'rb'))
	disjunct_modules = [x for x in custom_modules_lst if x not in python_modules]

	#This should be a good approximation of all external imports.
	return disjunct_modules
	
def search_conda_channel(conda_channel, elem, leave_me_alone=False):
	"""
	Search package within different conda channels.

	Parameters
	----------
	conda_channel : str
		conda channel to be searched
	elem : str
		package-name to be searched
	leave_me_alone : bool, optional
		if True most of the interactive queries are skipped, by default False

	Returns
	-------
	str: 
		package-name | None
	bool: 
		skip? True | False
	"""
	try:
		#This is a workaround as listening to a keyboard doesen't work on macos due to security reasons
		#Running the program via sudo feels wrong
		if leave_me_alone:
			print(f"Search {elem} within {conda_channel}...")	
		else:	
			cancel = input(f"Search {elem} within {conda_channel}?  press [enter] | [c] to cancel: ")
			if cancel == "c":
				print("Canceled search within this channel.")
				return None, True

		cmd = f'conda search -c {conda_channel} --json {elem}'.split()

		#TODO find better suitable parameters if there is time, currently only the seconds make sense 
		with tqdm(unit='s', unit_scale=True, miniters=1, mininterval=0, maxinterval=1, desc=elem) as t:
			#elapsed_time = time.time() - start_time
			call = subprocess.run(cmd,capture_output=True, shell=False)
			for line in call.stdout:
				t.set_description("processing...")
				t.update()
				sys.stdout.flush()

		call.check_returncode()
		print(f"Installation candidate for {elem} found!")

		output_dict = json.loads(call.stdout)
		output_lst = extract_keys(output_dict)
		installation_candidates = enumerate_installation_candidates(output_lst)

		if len(installation_candidates) == 1:
			# Return only available choice
			return installation_candidates[0], True

		else:
			print(f"Found several installation candidates for {elem}")
			choice = menu(installation_candidates, elem)

			print(f"choice is: {choice}")

			if choice == "s":
				print("Skip searching this channel...")
				return None, False
			elif choice == "c":
				return None, True
			elif choice == "n":
				print("Implement adding custom package name here..")
				package_name = choose_custom_name(elem)
				return package_name, True
			else:
				return installation_candidates[int(choice)], True

	except subprocess.CalledProcessError:
		print(f'{elem} not found within {conda_channel}.')
		if leave_me_alone:
			print("Trying next channel...")			
			return None, False
		else:
			choice = input("Do you want to insert a custom name instead? [y/n](default:no) | [c] for cancel :")
			if choice.lower() == "y":
				package_name = choose_custom_name(elem)
				return package_name, True
			elif choice.lower() == "c":
				return None, True
			else:
				print("Trying next channel...")			
				return None, False


def search_pip( elem, leave_me_alone=False):
	"""
	Search package within pip. 

	Notes:
		This functions gets only executed if the package wasn't found in neither
		of the provided conda-channels

		This regex failed to work after pip-search was deactivated - kept here in case of future usage.
		cleaned_lst = [re.split("(?P<version-number>\s\((\d+\.?\w*){1,3}\))", sample)[0] for sample in output]

	Parameters
	----------
	conda_channel : str
		conda channel to be searched
	elem : str
		package-name to be searched
	leave_me_alone : bool, optional
		if True most of the interactive queries are skipped, by default False

	Returns
	-------
	str: 
		package-name | None
	bool: 
		skip? True | False
	"""
	#This is a workaround as listening to a keyboard doesn't work on macos due to security reasons
	#Running the program via sudo feels wrong
	if leave_me_alone:
		print(f"Search {elem} within pip...")	
	else:	
		cancel = input(f"Search {elem} within pip?  press [enter] | [c] to cancel: ")
		if cancel == "c":
			#print("Cancel search. Skip to next package...")
			return None, True

	cmd = f'pip_search {elem}'.split()
	sp_return_code = subprocess.check_call(cmd, shell=False)
	
	sp = subprocess.run(cmd, shell=False, check=True,
						capture_output=True, text=True)

	output = sp.stdout.splitlines()
	cleaned_lst = [ sample.split(" ")[0] for sample in output][3:-1]
	installation_candidates = enumerate_installation_candidates(cleaned_lst)

	if len(installation_candidates) == 1:
		# Return only available choice
		return installation_candidates[0], True

	#Why the except block failed to execute? Also a change after switching to pip_search? Currently removed
	elif len(installation_candidates) == 0:
		return None, False

	else:
		print(f"Found several installation candidates for {elem}")
		choice = menu(installation_candidates, elem)

		if choice == "c" or choice == "s":
			return None, True
		elif choice == "n":
			print("Implement adding custom package name here..")
			package_name = choose_custom_name(elem)
			return package_name, True
		else:
			return installation_candidates[int(choice)], True

def create_env(conda_lst, pip_lst, env_name, channel_lst):
	"""
	Creates a valid environment.yml

	Parameters
	----------
	conda_lst : list
		list of packages to be installed via conda
	pip_lst : list
		list of packages to be installed via conda
	env_name : str
		name of the resulting environment.yml
	channel_lst : list
		list of conda-channels that are to be searched
	"""
	dependencies = conda_lst
	name = env_name
	
	if pip_lst:
		dependencies.append({'pip': pip_lst})


	dict_file = {'name': env_name,
				 'channels': channel_lst,
				 'dependencies': dependencies
				 }

	with open(f'{env_name}.yml', 'w') as file:
		print("Creating new environment.yml...")
		yaml.dump(dict_file, file, sort_keys=False)